from IPython.lib import guisupport
from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtconsole.inprocess import QtInProcessKernelManager


class QJupyterWidget(RichJupyterWidget):
    """Convenience class for a live Jupyter console widget."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        kernel_manager = QtInProcessKernelManager()
        kernel_manager.start_kernel()
        kernel = kernel_manager.kernel
        kernel.gui = "qt"

        kernel_client = kernel_manager.client()
        kernel_client.start_channels()

        self.kernel_manager = kernel_manager
        self.kernel_client = kernel_client

        def stop():
            kernel_client.stop_channels()
            kernel_manager.shutdown_kernel()
            guisupport.get_app_qt4().exit()

        self.exit_requested.connect(stop)

    def push_variables(self, variable_dict):
        """Given a dictionary containing name / value pairs,
        push those variables to the Jupyter console widget.
        """
        self.kernel_manager.kernel.shell.push(variable_dict)

    def clear(self):
        """Clears the terminal."""
        self._control.clear()

    def print_text(self, text):
        """Prints some plain text to the console."""
        self._append_plain_text(text)

    def execute_command(self, command):
        """Execute a command in the console widget."""
        self._execute(command, False)



import os
import sys
import numpy as np
from .qt import QtGui, QtWidgets, QtCore, Qt
import matplotlib.pyplot as plt
from .widgets import QJupyterWidget, MetaWidget, DataSetBrowser, DataSetPlotter
from .utils import load_json_ordered
import warnings

warnings.filterwarnings("ignore", message="The unit of the quantity is stripped")
warnings.filterwarnings("ignore", message="tight_layout")
warnings.filterwarnings("ignore", message="Tight layout not applied.")
warnings.filterwarnings("ignore", message="All-NaN slice encountered")
warnings.filterwarnings("ignore", message="All-NaN axis encountered")
warnings.filterwarnings("ignore", message="Attempting to set identical bottom")
warnings.filterwarnings("ignore", message="invalid value encountered in less")


class MainWindow(QtWidgets.QMainWindow):
    """Main window for scanning-squid-analysis gui."""

    def __init__(self):
        super().__init__()
        self.dataset = None
        self.setWindowTitle("ssm_analyze")
        self.shell = QJupyterWidget()
        self.station_snap = MetaWidget()
        self.measurement_meta = MetaWidget()
        self.dataset_browser = DataSetBrowser()
        self.dataset_browser.dataset_selector.clicked.connect(self.load_dataset)
        self.dataset_plotter = DataSetPlotter()
        self.dataset_browser.dataset_selector.clicked.connect(self.update_dataset_plot)
        self.shell.push_variables({"np": np, "plt": plt})

        self.file_menu = self.menuBar().addMenu("File")
        self.plot_menu = self.menuBar().addMenu("Plot")
        self.view_menu = self.menuBar().addMenu("View")

        self.dataset_dock = self.add_dock(
            self.dataset_browser, "DataSet Browser", "Left"
        )
        self.snapshot_dock = self.add_dock(
            self.station_snap, "Microscope Snapshot", "Left"
        )
        self.meta_dock = self.add_dock(
            self.measurement_meta, "Measurement Metadata", "Left"
        )
        self.shell_dock = self.add_dock(self.shell, "Shell", "Left")
        self.plotter_dock = self.add_dock(
            self.dataset_plotter, "DataSet Plotter", "Right"
        )
        self.tabifyDockWidget(self.snapshot_dock, self.meta_dock)

        self.plot_menu.addAction(
            "Export matplotlib...",
            self.dataset_plotter.export_mpl,
            QtGui.QKeySequence("Ctrl+P"),
        )
        # removing this feature due to a bug in pyqtgraph
        # see: https://github.com/pyqtgraph/pyqtgraph/issues/538
        # self.plot_menu.addAction('Export pyqtgraph...', self.dataset_plotter.export_qt,
        #                             QtGui.QKeySequence('Ctrl+Shift+P'))
        self.file_menu.addAction(
            "Select directory...",
            self.dataset_browser.select_from_dialog,
            QtGui.QKeySequence("Ctrl+O"),
        )
        self.file_menu.addAction(
            "Export current data...",
            self.dataset_plotter.export_data,
            QtGui.QKeySequence("Ctrl+S"),
        )

    def add_dock(self, widget, name, location, min_width=None):
        """Add a QDockWidget to the main window.
        Args:
            widget (QWidget): Widget to add to dock.
            name (str): Name to give dock.
            location (str): Where to put dock, in ('Left', 'Right', 'Top', 'Bottom').
            min_width (optional, int): Minimum width of the dock. Default: None.
        """
        dock = QtWidgets.QDockWidget(name)
        dock.setWidget(widget)
        self.view_menu.addAction(dock.toggleViewAction())
        loc_const = getattr(Qt, f"{location}DockWidgetArea")
        self.addDockWidget(loc_const, dock)
        if min_width is not None:
            dock.setMinimumWidth(min_width)
        return dock

    def load_dataset(self):
        """Get dataset selected in self.dataset_browser."""
        dataset = self.dataset_browser.get_dataset()
        if "snapshot.json" not in os.listdir(dataset.location):
            return
        path = os.path.join(dataset.location, "snapshot.json")
        meta = load_json_ordered(path)
        snap = meta.pop("station")
        self.station_snap.load_meta(data=snap)
        self.measurement_meta.load_meta(data=meta)
        self.dataset = dataset
        self.shell.push_variables({"dataset": self.dataset})

    def update_dataset_plot(self):
        """Get dataset and update plot."""
        self.load_dataset()
        if isinstance(self.dataset, str):
            self.dataset_plotter.update(self.dataset)
            return
        if self.dataset is None or "snapshot.json" not in os.listdir(
            self.dataset.location
        ):
            return
        self.dataset_plotter.update(self.dataset)
        self.shell.push_variables({"arrays": self.dataset_plotter.arrays})


def main():
    icon_path = os.path.join(os.path.dirname(__file__), "img", "icon.png")
    app = QtWidgets.QApplication([])
    app.setWindowIcon(QtGui.QIcon(icon_path))

    win = MainWindow()
    sample_data_dir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), os.pardir, "sample_data", "data")
    )
    win.dataset_browser.directory = sample_data_dir
    win.dataset_browser.set_available_dates()
    win.showMaximized()

    app.lastWindowClosed.connect(sys.exit)
    app.exec_()


if __name__ == "__main__":
    main()



from typing import Dict, Sequence, Any, Optional
import numpy as np
from pint import UnitRegistry
from collections import OrderedDict
import json


def load_json_ordered(filename: str) -> OrderedDict:
    """Loads json file as an ordered dict.
    Args:
        filename: Path to json file to be loaded.
    Returns:
        OrderedDict: odict
            OrderedDict containing data from json file.
    """
    with open(filename) as f:
        odict = json.load(f, object_pairs_hook=OrderedDict)
    return odict


def set_h5_attrs(g, kwargs):
    """Adds data to HDF5 group `g` from dict `kwargs`."""
    for name, value in kwargs.items():
        if isinstance(value, dict):
            sub_g = g.create_group(name)
            set_h5_attrs(sub_g, value)
        else:
            if isinstance(value, np.ndarray):
                g.create_dataset(name, data=value)
            else:
                g.attrs[name] = value


def make_scan_vectors(
    scan_params: Dict[str, Any], ureg: Any
) -> Dict[str, Sequence[float]]:
    """Creates x and y vectors for given scan parameters.

    Args:
        scan_params: Scan parameter dict
        ureg: pint UnitRegistry, manages units.

    Returns:
        Dict: scan_vectors
            {axis_name: axis_vector} for x, y axes.
    """
    Q_ = ureg.Quantity
    center = []
    size = []
    rng = []
    for ax in ["x", "y"]:
        center.append(Q_(scan_params["center"][ax]).to("V").magnitude)
        size.append(scan_params["scan_size"][ax])
        rng.append(Q_(scan_params["range"][ax]).to("V").magnitude)
    x = np.linspace(center[0] - 0.5 * rng[0], center[0] + 0.5 * rng[0], size[0])
    y = np.linspace(center[1] - 0.5 * rng[1], center[1] + 0.5 * rng[1], size[1])
    return {"x": x, "y": y}


def make_xy_grids(
    scan_vectors: Dict[str, Sequence[float]], slow_ax: str, fast_ax: str
) -> Dict[str, Any]:
    """Makes meshgrids from x, y scan_vectors (used for plotting, etc.).

    Args:
        scan_vectors: Dict of {axis_name: axis_vector} for x, y axes (from make_scan_vectors).
        slow_ax: Name of scan slow axis ('x' or 'y').
        fast_ax: Name of scan fast axis ('x' or 'y').

    Returns:
        Dict: xy_grids
            {axis_name: axis_grid} for x, y axes.
    """
    slow_ax_vec = scan_vectors[slow_ax]
    fast_ax_vec = scan_vectors[fast_ax]
    if fast_ax == "y":
        X, Y = np.meshgrid(slow_ax_vec, fast_ax_vec, indexing="ij")
    else:
        X, Y = np.meshgrid(fast_ax_vec, slow_ax_vec, indexing="xy")
    return {"x": X, "y": Y}


def to_real_units(data_set: Any, ureg: Any = None) -> Any:
    """Converts DataSet arrays from DAQ voltage to real units using recorded metadata.
        Preserves shape of DataSet arrays.

    Args:
        data_set: qcodes DataSet created by Microscope.scan_plane
        ureg: Pint UnitRegistry. Default None.

    Returns:
        np.ndarray: data
            ndarray like the DataSet array, but in real units as prescribed by
            factors in DataSet metadata.
    """
    if ureg is None:
        from pint import UnitRegistry

        ureg = UnitRegistry()
        ureg.define("Phi0 = Phi_0")
        ureg.define("Ohm = ohm")
    meta = data_set.metadata["loop"]["metadata"]
    data = np.full_like(data_set.daq_ai_voltage, np.nan, dtype=np.double)
    for i, ch in enumerate(meta["channels"].keys()):
        array = data_set.daq_ai_voltage[:, i, :] * ureg("V")
        unit = meta["channels"][ch]["unit"]
        data[:, i, :] = (array * ureg.Quantity(meta["prefactors"][ch])).to(unit)
    return data


def scan_to_arrays(
    scan_data: Any,
    ureg: Optional[Any] = None,
    real_units: Optional[bool] = True,
    xy_unit: Optional[str] = None,
) -> Dict[str, Any]:
    """Extracts scan data from DataSet and converts to requested units.

    Args:
        scan_data: qcodes DataSet created by Microscope.scan_plane
        ureg: pint UnitRegistry, manages physical units.
        real_units: If True, converts z-axis data from DAQ voltage into
            units specified in measurement configuration file.
        xy_unit: String describing quantity with dimensions of length.
            If xy_unit is not None, scanner x, y DAQ ao voltage will be converted to xy_unit
            according to scanner constants defined in microscope configuration file.

    Returns:
        Dict: arrays
            Dict of x, y vectors and grids, and measured data in requested units.
    """
    if ureg is None:
        from pint import UnitRegistry

        ureg = UnitRegistry()
        ureg.define("Phi0 = Phi_0")
        ureg.define("Ohm = ohm")
    Q_ = ureg.Quantity
    meta = scan_data.metadata["loop"]["metadata"]
    scan_vectors = make_scan_vectors(meta, ureg)
    slow_ax = "x" if meta["fast_ax"] == "y" else "y"
    grids = make_xy_grids(scan_vectors, slow_ax, meta["fast_ax"])
    arrays = {"X": grids["x"] * ureg("V"), "Y": grids["y"] * ureg("V")}
    arrays.update(
        {"x": scan_vectors["x"] * ureg("V"), "y": scan_vectors["y"] * ureg("V")}
    )
    for ch, info in meta["channels"].items():
        array = scan_data.daq_ai_voltage[:, info["idx"], :] * ureg("V")
        if meta["fast_ax"] == "y":
            array = array.T
        if real_units:
            pre = meta["prefactors"][ch]
            arrays.update({ch: (Q_(pre) * array).to(info["unit"])})
        else:
            arrays.update({ch: array})
    if real_units and xy_unit is not None:
        bendc = scan_data.metadata["station"]["instruments"]["benders"]["metadata"][
            "constants"
        ]
        for ax in ["x", "y"]:
            grid = (grids[ax] * ureg("V") * Q_(bendc[ax])).to(xy_unit)
            vector = (scan_vectors[ax] * ureg("V") * Q_(bendc[ax])).to(xy_unit)
            arrays.update({ax.upper(): grid, ax: vector})
    return arrays


def td_to_arrays(
    td_data: Any,
    ureg: Optional[Any] = None,
    real_units: Optional[bool] = True,
    z_unit: Optional[str] = None,
) -> Dict[str, Any]:
    """Extracts scan data from DataSet and converts to requested units.

    Args:
        td_data: qcodes DataSet created by Microscope.td_cap
        ureg: pint UnitRegistry, manages physical units.
        real_units: If True, converts data from DAQ voltage into
            units specified in measurement configuration file.
        z_unit: String describing quantity with dimensions of length.
            If z_unit is not None, scanner z DAQ ao voltage will be converted to z_unit
            according to scanner constant defined in microscope configuration file.
    Returns:
        Dict: arrays
            Dict of measured data in requested units.
    """
    if ureg is None:
        from pint import UnitRegistry

        ureg = UnitRegistry()
        ureg.define("Phi0 = Phi_0")
        ureg.define("Ohm = ohm")
    Q_ = ureg.Quantity
    meta = td_data.metadata["loop"]["metadata"]
    h = [Q_(val).to("V").magnitude for val in meta["range"]]
    dV = Q_(meta["dV"]).to("V").magnitude
    heights = np.linspace(h[0], h[1], int((h[1] - h[0]) / dV)) * ureg("V")
    arrays = {}
    # arrays = {'height': heights * ureg('V')}
    for ch, info in meta["channels"].items():
        array = td_data.daq_ai_voltage[:, info["idx"], 0] * ureg("V")
        last_idx = min(len(heights), len(array[np.isfinite(array)]))
        if real_units:
            pre = meta["prefactors"][ch]
            arrays.update({ch: (Q_(pre) * array[:last_idx]).to(info["unit"])})
        else:
            arrays.update({ch: array[:last_idx]})
    if real_units and z_unit is not None:
        bendc = td_data.metadata["station"]["instruments"]["benders"]["metadata"][
            "constants"
        ]
        heights = (heights * Q_(bendc["z"])).to(z_unit)
    arrays.update({"height": heights[:last_idx]})
    return arrays



from ..qt import QtGui, QtWidgets


class ItemComboBox(QtWidgets.QComboBox):
    def go_to_item(self, item):
        idx = self.findText(item)
        if idx >= 0:
            self.setCurrentIndex(idx)
        else:
            raise ValueError("No item %d" % idx)


class NoEditItem(QtGui.QStandardItem):
    def __init__(self, name):
        super(NoEditItem, self).__init__(name)
        self.setEditable(False)

    def findItems(self, name):
        matches = []
        for i in range(self.rowCount()):
            item = self.child(i)
            if str(item.text()) == name:
                matches.append(item)
        return matches



from matplotlib.figure import Figure
from matplotlib.collections import LineCollection
from matplotlib import cm
from matplotlib import colors
from matplotlib import transforms
import matplotlib.pyplot as plt

plt.style.use("seaborn")

from matplotlib.widgets import Slider
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.gridspec as gridspec
import numpy as np
from scipy.linalg import lstsq
from scipy.ndimage.interpolation import rotate
from scipy.io import savemat
import h5py
import pyqtgraph.exporters
import pyqtgraph as pg
import pickle
from ..qt import *
from .sliders import SliderWidget, VertSlider
from ..utils import *

mpl_cmaps = ("viridis", "plasma", "inferno", "magma", "cividis", "Greys")
qt_cmaps = (
    "thermal",
    "flame",
    "yellowy",
    "bipolar",
    "grey",
)  # , 'spectrum', 'cyclic', 'greyclip')
plot_lw = 3
font_size = 12
plt.rcParams.update({"font.size": font_size})


__all__ = ["DataSetPlotter"]


def format_units(units, skip=True):
    if skip:
        return str(units)
    units = str(units).lower()
    scale_map = {
        "giga": "G",
        "mega": "M",
        "kilo": "k",
        "milli": "m",
        "micro": "$\\mu$",
        "nano": "n",
        "pico": "p",
        "femto": "f",
    }
    unit_map = {
        "meter": "m",
        "phi0": "$\\Phi_0$",
        "farad": "F",
        "ampere": "A",
        "volt": "V",
    }
    for name, abbrev in scale_map.items():
        units = units.replace(name, abbrev)
    for name, abbrev in unit_map.items():
        units = units.replace(name, abbrev)
    return units


class PlotWidget(QtWidgets.QWidget):
    """Plotting widget comprised of a matplotlib canvas and a pyqtgraph widget, along with some
    options like slices of image data, axis transforms, etc.
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.current_data = None
        self.exp_data = {}

        self.fig = Figure()
        self.fig.patch.set_alpha(1)
        self.fig.subplots_adjust(bottom=0.15)
        self.canvas = FigureCanvasQTAgg(self.fig)
        self.canvas.setParent(self)
        self.canvas.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        self.toolbar = NavigationToolbar2QT(self.canvas, self)

        self.pyqt_plot = SlicePlotWidget(parent=self)
        self.pyqt_imview = SliceableImageView(parent=self)
        self.pyqt_plot.hide()
        self.pyqt_imview.hide()
        pyqt_top_widgets = QtWidgets.QWidget(parent=self)
        pyqt_bottom_widgets = QtWidgets.QWidget(parent=self)
        pyqt_top_layout = QtWidgets.QVBoxLayout(pyqt_top_widgets)
        pyqt_bottom_layout = QtWidgets.QVBoxLayout(pyqt_bottom_widgets)
        pyqt_top_layout.addWidget(self.pyqt_plot)
        pyqt_top_layout.addWidget(self.pyqt_imview)
        pyqt_bottom_layout.addWidget(self.pyqt_imview.x_slice_widget)
        pyqt_bottom_layout.addWidget(self.pyqt_imview.y_slice_widget)
        self.pyqt_imview.x_slice_widget.hide()
        self.pyqt_imview.y_slice_widget.hide()
        self.pyqt_splitter = QtWidgets.QSplitter(Qt.Vertical, parent=self)
        self.pyqt_splitter.hide()
        self.pyqt_splitter.addWidget(pyqt_top_widgets)
        self.pyqt_splitter.addWidget(pyqt_bottom_widgets)
        self.pyqt_splitter.setStretchFactor(1, 1.5)

        self.option_layout = QtWidgets.QHBoxLayout()
        self._setup_cmap()
        self._setup_background_subtraction()
        self._setup_transforms()
        self._setup_slices()
        self._setup_options()
        self.slice_state = 1  # no slices
        self.set_slice()

        layout = QtWidgets.QVBoxLayout(self)
        layout.addLayout(self.option_layout)
        layout.addWidget(self.backsub_widget)
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        layout.addWidget(self.pyqt_splitter)
        layout.addWidget(self.rotate_widget)

    def _setup_cmap(self):
        """Setup the UI for selecting matplotlib and pyqtgraph colormaps."""
        self.mpl_cmap = "viridis"
        self.mpl_cmap_selector = QtWidgets.QComboBox()
        self.mpl_cmap_selector.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.qt_cmap_selector = QtWidgets.QComboBox()
        self.qt_cmap_selector.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)

        cmap_widget = QtWidgets.QGroupBox("Colormaps")
        mpl_cmap_widget = QtWidgets.QGroupBox("matplotlib")
        mpl_cmap_layout = QtWidgets.QHBoxLayout(mpl_cmap_widget)
        mpl_cmap_layout.addWidget(self.mpl_cmap_selector)
        qt_cmap_widget = QtWidgets.QGroupBox("pyqtgraph")
        qt_cmap_layout = QtWidgets.QHBoxLayout(qt_cmap_widget)
        qt_cmap_layout.addWidget(self.qt_cmap_selector)
        cmap_layout = QtWidgets.QHBoxLayout(cmap_widget)
        mpl_cmap_widget.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Minimum
        )
        qt_cmap_widget.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Minimum
        )
        cmap_layout.addWidget(mpl_cmap_widget)
        cmap_layout.addWidget(qt_cmap_widget)
        self.option_layout.addWidget(cmap_widget)

        self.mpl_cmap_selector.currentIndexChanged.connect(self.set_cmap_mpl)
        for name in mpl_cmaps:
            self.mpl_cmap_selector.addItem(name)
        self.mpl_cmap_selector.setCurrentIndex(0)
        self.qt_cmap_selector.currentIndexChanged.connect(self.set_cmap_qt)
        for name in qt_cmaps:
            self.qt_cmap_selector.addItem(name)
        self.qt_cmap_selector.setCurrentIndex(0)

    def _setup_background_subtraction(self):
        """Setup UI for global or line-by-line background subtraction."""
        self.backsub_radio = QtWidgets.QButtonGroup()
        backsub_buttons = [
            QtWidgets.QRadioButton(s)
            for s in ("none", "min", "max", "mean", "median", "linear")
        ]
        backsub_buttons[0].setChecked(True)
        self.backsub_widget = QtWidgets.QGroupBox("Background subtraction")
        self.backsub_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum
        )
        backsub_layout = QtWidgets.QHBoxLayout(self.backsub_widget)
        for i, b in enumerate(backsub_buttons):
            backsub_layout.addWidget(b)
            self.backsub_radio.addButton(b, i)
        self.backsub_radio.buttonClicked.connect(self.replot)

        self.line_backsub_radio = QtWidgets.QButtonGroup()
        self.line_backsub_btn = QtWidgets.QCheckBox("line-by-line")
        self.line_backsub_btn.setChecked(False)
        self.x_line_backsub_btn, self.y_line_backsub_btn = xy_btns = [
            QtWidgets.QRadioButton(s) for s in ("x", "y")
        ]
        xy_btns[0].setChecked(True)
        backsub_layout.addWidget(self.line_backsub_btn)
        for i, b in enumerate(xy_btns):
            b.setEnabled(False)
            backsub_layout.addWidget(b)
            self.line_backsub_radio.addButton(b, i)

        self.line_backsub_btn.stateChanged.connect(self.update_line_by_line)
        self.line_backsub_radio.buttonClicked.connect(self.replot)

    def _setup_transforms(self):
        """Setup UI for axis transformations. Currently this is only rotation.
        TODO: Add flipud/fliplr?
        """
        self.rotate_widget = QtWidgets.QGroupBox("Rotate")
        rotate_layout = QtWidgets.QVBoxLayout()
        self.rotate_widget.setLayout(rotate_layout)
        self.rotate_slider = SliderWidget(-180, 180, 0, 60)
        rotate_layout.addWidget(self.rotate_slider)
        self.rotate_slider.value_box.valueChanged.connect(self.replot)

    def _setup_slices(self):
        """Setup UI for slices of image/2D data."""
        self.slice_radio = QtWidgets.QButtonGroup()
        slice_buttons = [QtWidgets.QRadioButton(s) for s in ("none", "x", "y")]
        slice_buttons[0].setChecked(True)
        slice_widget = QtWidgets.QGroupBox("Slice")
        slice_layout = QtWidgets.QVBoxLayout(slice_widget)
        for i, b in enumerate(slice_buttons):
            slice_layout.addWidget(b)
            self.slice_radio.addButton(b, i)
        self.line_color_btn = QtWidgets.QCheckBox("color by value")
        self.line_color_btn.setChecked(True)
        self.line_color_btn.stateChanged.connect(self.replot)
        slice_layout.addWidget(self.line_color_btn)
        self.option_layout.addWidget(slice_widget)
        self.slice_radio.buttonClicked.connect(self.set_slice)
        slice_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum
        )

    def _setup_options(self):
        """Setup UI for other plot options."""
        opt_group = QtWidgets.QGroupBox("Plot options")
        opt_layout = QtWidgets.QVBoxLayout()
        opt_group.setLayout(opt_layout)
        plot_opts = [
            ("pyqtgraph", False),
            ("grid", True),
            ("histogram", False),
        ]
        self.opt_checks = {}
        for name, checked in plot_opts:
            btn = QtWidgets.QCheckBox(name)
            btn.setChecked(checked)
            opt_layout.addWidget(btn)
            btn.stateChanged.connect(self.replot)
            self.opt_checks[name] = btn
        self.bins_box = QtWidgets.QSpinBox()
        self.bins_box.setMinimum(10)
        self.bins_box.setMaximum(1000)
        self.bins_box.setValue(100)
        self.bins_box.setKeyboardTracking(False)
        self.bins_box.valueChanged.connect(self.replot)
        self.bins_box.setEnabled(self.get_opt("histogram"))

        bins_widget = QtWidgets.QWidget()
        bins_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum
        )
        bins_layout = QtWidgets.QHBoxLayout(bins_widget)
        if self.get_opt("histogram"):
            bins_widget.show()
        else:
            bins_widget.hide()
        bins_layout.addWidget(QtWidgets.QLabel("bins: "))
        bins_layout.addWidget(self.bins_box)
        opt_layout.addWidget(bins_widget)

        # opt_layout.addWidget(QtWidgets.QLabel('bins:'))
        # opt_layout.addWidget(self.bins_box)
        self.option_layout.addWidget(opt_group)
        self.opt_checks["histogram"].stateChanged.connect(
            lambda val: self.bins_box.setEnabled(val)
        )
        self.opt_checks["histogram"].stateChanged.connect(
            lambda val: bins_widget.show() if val else bins_widget.hide()
        )

    def set_cmap_mpl(self, idx):
        """Set the matplotlib colormap.
        Args:
            idx (int): Index of the requested colormap in self.mpl_cmap_selector.
        """
        name = str(self.mpl_cmap_selector.itemText(idx))
        if not name:
            return
        self.mpl_cmap = name
        self.replot()

    def set_cmap_qt(self, idx):
        """Set the pyqtgraph colormap.
        Args:
            idx (int): Index of the requested colormap in self.qt_cmap_selector.
        """
        name = str(self.qt_cmap_selector.itemText(idx))
        if not name:
            return
        self.pyqt_imview.set_cmap(name)

    def get_opt(self, optname):
        """Returns True if option `optname` is checked, else False.
        Args:
            optname (str): Name of plot option to check.
        """
        return bool(self.opt_checks[optname].isChecked())

    def plot_arrays(self, xs, ys, zs=None, title=""):
        """Plots data based on dimension and all user-selected options, transforms, etc.
        Args:
            xs (list[str, np.ndarray[pint.Quantity]]): x data in the form of a
                list of [name, 0D or 1D array of pint.Quantities].
            ys (list[str, np.ndarray[pint.Quantity]]): y data in the form of a
                list of [name, 0D or 1D array of pint.Quantities].
            zs (optional, list[str, np.ndarray[pint.Quantity]]): None if plotting 1D data, else
                z data in the form of a list of [name, 2D array of pint.Quantities]. Default: None.
            title (optional, str): Title for matplotlib figure. Default: ''.
        """
        self.fig_title = title
        self.fig.clear()
        self.pyqt_plot.clear()
        self.toolbar.hide()
        self.canvas.hide()
        self.pyqt_splitter.hide()
        self.pyqt_plot.hide()
        self.pyqt_imview.hide()
        if zs is None:  # 1d data
            self.line_backsub_btn.setChecked(False)
            self.line_backsub_btn.setEnabled(False)
            self.pyqt_imview.x_slice_widget.hide()
            self.pyqt_imview.y_slice_widget.hide()
            self.plot_1d(xs, ys)
        else:  # 2d data
            self.line_backsub_btn.setEnabled(True)
            angle = self.rotate_slider.value_box.value()  # degrees
            if self.slice_state == 0:
                slice_state = None
            elif self.slice_state == 1:
                slice_state = "x"
                self.pyqt_imview.x_slice_widget.show()
                self.pyqt_imview.y_slice_widget.hide()
            elif self.slice_state == 2:
                slice_state = "y"
                self.pyqt_imview.x_slice_widget.hide()
                self.pyqt_imview.y_slice_widget.show()
            self.plot_2d(xs, ys, zs, angle=angle, slice_state=slice_state)
        self.fig.suptitle(self.fig_title, fontsize=12)
        self.canvas.draw()

    def plot_1d(self, xs, ys):
        """Plot 1D data according to user-selected options, transformations, etc.
        Args:
            xs (list[str, np.ndarray[pint.Quantity]]): x data in the form of a
                list of [name, 0D or 1D array of pint.Quantities].
            ys (list[str, np.ndarray[pint.Quantity]]): y data in the form of a
                list of [name, 0D or 1D array of pint.Quantities].
        """
        label = ys[0]
        use_pyqt = self.get_opt("pyqtgraph")
        xlabel = f"{xs[0]} [{format_units(xs[1].units, use_pyqt)}]"
        ylabel = f"{ys[0]} [{format_units(ys[1].units, use_pyqt)}]"
        marker = "."
        # ymin, ymax = np.min(ys[1]), np.max(ys[1])
        if use_pyqt:
            self.plot_1d_qt(xs, ys, xlabel, ylabel, label)
            self.pyqt_plot.show()
            self.pyqt_splitter.show()
        else:
            self.plot_1d_mpl(xs, ys, xlabel, ylabel, marker, label)
            self.toolbar.show()
            self.canvas.show()
        self.rotate_widget.hide()
        self.exp_data = {
            d[0]: {"array": d[1].magnitude, "unit": str(d[1].units)} for d in (xs, ys)
        }

    def plot_1d_qt(self, xs, ys, xlabel, ylabel, label):
        """Plot 1D data on self.pyqt_plot.
        Args:
            xs (list[str, np.ndarray[pint.Quantity]]): x data in the form of a
                list of [name, 0D or 1D array of pint.Quantities].
            ys (list[str, np.ndarray[pint.Quantity]]): y data in the form of a
                list of [name, 0D or 1D array of pint.Quantities].
            xlabel (str): x-axis label.
            ylabel (str): y-axis label.
            label (str): Label for legend.
        """
        self.pyqt_plot.setLabels(bottom=(xlabel,), left=(ylabel,))
        self.pyqt_plot.plot(xs[1].magnitude, ys[1].magnitude, symbol="o", pen=None)
        grid = self.get_opt("grid")
        self.pyqt_plot.plotItem.showGrid(x=grid, y=grid)

    def plot_1d_mpl(self, xs, ys, xlabel, ylabel, marker, label):
        """Plot 1D data on self.fig.
        Args:
            xs (list[str, np.ndarray[pint.Quantity]]): x data in the form of a
                list of [name, 0D or 1D array of pint.Quantities].
            ys (list[str, np.ndarray[pint.Quantity]]): y data in the form of a
                list of [name, 0D or 1D array of pint.Quantities].
            xlabel (str): x-axis label.
            ylabel (str): y-axis label.
            marker (str): Plot point marker.
            label (str): Label for legend.
        """
        ax = self.fig.add_subplot(111)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.plot(xs[1].magnitude, ys[1].magnitude, marker, label=label)
        ax.grid(self.get_opt("grid"))
        self.fig.tight_layout()
        self.fig.subplots_adjust(top=0.9, bottom=0.15)
        ax.legend()

    def plot_2d(self, xs, ys, zs, cmap=None, angle=0, slice_state=None):
        """Plot 2D data according to user-selected options, transformations, etc.
        Args:
            xs (list[str, np.ndarray[pint.Quantity]]): x data in the form of a
                list of [name, 0D or 1D array of pint.Quantities).
            ys (list[str, np.ndarray[pint.Quantity]]): y data in the form of a
                list of [name, 0D or 1D array of pint.Quantities).
            zs (list[str, np.ndarray[pint.Quantity]]): z data in the form of a list of
                [name, 2D array of pint.Quantities).
            cmap (optional, str): Name of matplotlib colormap. Default: None.
            angle (optional, float): Angle by which to rotate image (degrees). Default: 0.
            slice_state (optional, str): Requested 1D slice state, in (None, 'x', 'y'). Default: None.
        """
        cmap = cmap or self.mpl_cmap
        use_pyqt = self.get_opt("pyqtgraph")
        xlabel = f"{xs[0]} [{format_units(xs[1].units, use_pyqt)}]"
        ylabel = f"{ys[0]} [{format_units(ys[1].units, use_pyqt)}]"
        zlabel = f"{zs[0]} [{format_units(zs[1].units, use_pyqt)}]"
        zmin, zmax = np.nanmin(zs[1].magnitude), np.nanmax(zs[1].magnitude)
        self.rotate_widget.show()
        if use_pyqt:
            self.plot_2d_qt(xs, ys, zs, xlabel, ylabel, zlabel, angle=angle)
            self.pyqt_imview.show()
            self.pyqt_splitter.show()
        else:
            self.plot_2d_mpl(
                xs,
                ys,
                zs,
                xlabel,
                ylabel,
                zlabel,
                vmin=zmin,
                vmax=zmax,
                cmap=cmap,
                angle=angle,
                slice_state=slice_state,
            )
            self.toolbar.show()
            self.canvas.show()

    def plot_2d_qt(self, xs, ys, zs, xlabel, ylabel, zlabel, angle=0):
        """Plot 2D data on self.pyqt_imview.
        Args:
            xs (list[str, np.ndarray[pint.Quantity]]): x data in the form of a
                list of [name, 0D or 1D array of pint.Quantities).
            ys (list[str, np.ndarray[pint.Quantity]]): y data in the form of a
                list of [name, 0D or 1D array of pint.Quantities).
            zs (list[str, np.ndarray[pint.Quantity]]): z data in the form of a list of
                [name, 2D array of pint.Quantities).
            xlabel (str): x-axis label.
            ylabel (str): y-axis label.
            zlabel (str): z-axis label.
            angle (optional, float): Angle by which to rotate image (degrees). Default: 0.
        """
        pos = np.nanmin(xs[1][0].magnitude), np.nanmin(ys[1][0].magnitude)
        scale = (
            np.ptp(xs[1].magnitude) / zs[1].shape[0],
            np.ptp(ys[1].magnitude) / zs[1].shape[1],
        )
        z = rotate(zs[1].magnitude.T, angle, cval=np.nanmin(zs[1].magnitude))
        self.pyqt_imview.setImage(z, pos=pos, scale=scale)
        self.pyqt_imview.setLabels(xlabel=xlabel, ylabel=ylabel, zlabel=zlabel)
        self.pyqt_imview.autoRange()
        self.pyqt_imview.set_histogram(self.get_opt("histogram"))
        grid = self.get_opt("grid")
        self.pyqt_imview.getView().showGrid(grid, grid)
        self.pyqt_imview.x_slice_widget.plotItem.showGrid(x=grid, y=grid)
        self.pyqt_imview.y_slice_widget.plotItem.showGrid(x=grid, y=grid)
        self.exp_data = {
            d[0]: {"array": d[1].magnitude, "unit": str(d[1].units)} for d in (xs, ys)
        }
        self.exp_data[zs[0]] = {"array": z, "unit": str(zs[1].units)}

    def plot_2d_mpl(
        self,
        xs,
        ys,
        zs,
        xlabel,
        ylabel,
        zlabel,
        cmap=None,
        angle=0,
        slice_state=None,
        **kwargs,
    ):
        """Plot 2D data on self.fig, with options determined by keyword args.
        Args:
            xs (list[str, np.ndarray[pint.Quantity]]): x data in the form of a
                list of [name, 0D or 1D array of pint.Quantities).
            ys (list[str, np.ndarray[pint.Quantity]]): y data in the form of a
                list of [name, 0D or 1D array of pint.Quantities).
            zs (list[str, np.ndarray[pint.Quantity]]): z data in the form of a list of
                [name, 2D array of pint.Quantities).
            xlabel (str): x-axis label.
            ylabel (str): y-axis label.
            zlabel (str): z-axis label.
            cmap (optional, str): Name of matplotlib colormap. Default: None.
            angle (optional, float): Angle by which to rotate image (degrees). Default: 0.
            slice_state (optional, str): Requested 1D slice state, in (None, 'x', 'y'). Default: None.
            kwargs (optional, dict): Keyword arguments passed to plt.pcolormesh constructor.
        """
        if slice_state is None:
            plt.rcParams.update({"font.size": font_size})
            self.fig.subplots_adjust(
                top=0.9, bottom=0.15, left=0.0, right=1, hspace=0.0, wspace=0
            )
            ax = self.fig.add_subplot(111)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            x0, y0 = np.mean(xs[1].magnitude), np.mean(ys[1].magnitude)
            tr = transforms.Affine2D().rotate_deg_around(x0, y0, angle)
            im = ax.pcolormesh(
                xs[1].magnitude,
                ys[1].magnitude,
                zs[1].magnitude,
                cmap=cmap,
                shading="auto",
                transform=(tr + ax.transData),
                **kwargs,
            )
            main_divider = make_axes_locatable(ax)
            cax = main_divider.append_axes("right", size="10%", pad=0.2)
            cbar = plt.colorbar(im, cax=cax)
            cbar.set_label(zlabel)
            ax.set_aspect("equal")
            if self.get_opt("histogram"):
                nbins = self.bins_box.value()
                min_val, max_val = np.nanmin(zs[1].magnitude), np.nanmax(
                    zs[1].magnitude
                )
                # add axis for histogram
                ax_hist = main_divider.append_axes("right", size="25%", pad=1.1)
                # lines indicating cmin and cmax on histogram
                upper = ax_hist.axhline(max_val, color="k", lw=2)
                lower = ax_hist.axhline(min_val, color="k", lw=2)
                ax_hist.set_ylim(cax.get_xlim())
                ax_hist.set_xticklabels([])
                ax_hist.grid(self.get_opt("grid"))
                # ax_hist.invert_xaxis()
                N, bins, patches = ax_hist.hist(
                    zs[1].magnitude.ravel(), bins=nbins, orientation="horizontal"
                )
                # set color of histogram bins according to z value
                fracs = np.linspace(min_val, max_val, nbins)
                norm = colors.Normalize(min_val, max_val)
                for frac, patch in zip(fracs, patches):
                    color = getattr(cm, cmap)(norm(frac))
                    patch.set_facecolor(color)
                # make sliders to control cmin and cmax
                ax_min_slider = main_divider.append_axes("right", size="10%", pad=0.3)
                self.min_slider = min_slider = VertSlider(
                    ax_min_slider,
                    "min",
                    min_val,
                    max_val,
                    fontsize=font_size,
                    valinit=min_val,
                    labels=True,
                    alpha=1,
                    facecolor=getattr(cm, cmap)(norm(min_val)),
                )
                ax_max_slider = main_divider.append_axes("right", size="10%", pad=0.3)
                self.max_slider = max_slider = VertSlider(
                    ax_max_slider,
                    "max",
                    min_val,
                    max_val,
                    slidermin=min_slider,
                    fontsize=font_size,
                    valinit=max_val,
                    labels=True,
                    alpha=1,
                    facecolor=getattr(cm, cmap)(norm(max_val)),
                    start_at_bottom=False,
                )
                # function called when min_slider or max_slider are changed
                def update_cval(val):
                    cmin = min_slider.val
                    cmax = max_slider.val
                    cbar.mappable.set_clim([cmin, cmax])
                    im.set_clim([cmin, cmax])
                    upper.set_ydata(cmax)
                    lower.set_ydata(cmin)
                    min_slider.valmax = cmax
                    # update colors on the histogram to reflect current clims
                    norm = colors.Normalize(cmin, cmax)
                    for frac, patch in zip(fracs, patches):
                        if cmin <= frac <= cmax:
                            color = getattr(cm, cmap)(norm(frac))
                            patch.set_alpha(1)
                        else:
                            color = "k"
                            patch.set_alpha(0.25)
                        patch.set_facecolor(color)

                for s in [min_slider, max_slider]:
                    s.on_changed(update_cval)
            z = rotate(zs[1].magnitude.T, angle, cval=np.nan)
            x = np.linspace(*ax.get_xlim(), z.shape[1])
            y = np.linspace(*ax.get_ylim(), z.shape[0])
            self.exp_data = {
                xs[0]: {"array": x, "unit": str(xs[1].units)},
                ys[0]: {"array": y, "unit": str(ys[1].units)},
                zs[0]: {"array": z.T, "unit": str(zs[1].units)},
            }
        else:  # slicing
            plt.rcParams.update({"font.size": 10})
            self.fig.subplots_adjust(
                top=0.85, bottom=0.05, left=0.0, right=1.0, hspace=0.5, wspace=0.0
            )
            ax0 = plt.subplot2grid((12, 12), (0, 3), colspan=6, rowspan=5, fig=self.fig)
            ax0.set_xlabel(xlabel)
            ax0.set_ylabel(ylabel)
            x0, y0 = np.mean(xs[1].magnitude), np.mean(ys[1].magnitude)
            tr = transforms.Affine2D().rotate_deg_around(x0, y0, angle)
            im = ax0.pcolormesh(
                xs[1],
                ys[1],
                zs[1].magnitude,
                cmap=cmap,
                shading="auto",
                transform=(tr + ax0.transData),
                **kwargs,
            )
            main_divider = make_axes_locatable(ax0)
            cax = main_divider.append_axes("right", size="10%", pad=0.2)
            cbar = plt.colorbar(im, cax=cax)
            cbar.set_label(zlabel)
            ax0.set_aspect("equal")
            if self.get_opt("histogram"):
                nbins = self.bins_box.value()
                min_val, max_val = np.nanmin(zs[1].magnitude), np.nanmax(
                    zs[1].magnitude
                )
                # add axis for histogram
                ax_hist = main_divider.append_axes("right", size="50%", pad=1.1)
                # lines indicating cmin and cmax on histogram
                upper = ax_hist.axhline(max_val, color="k", lw=2)
                lower = ax_hist.axhline(min_val, color="k", lw=2)
                ax_hist.set_ylim(cax.get_xlim())
                ax_hist.set_xticklabels([])
                ax_hist.grid(self.get_opt("grid"))
                # ax_hist.invert_xaxis()
                N, bins, patches = ax_hist.hist(
                    zs[1].magnitude.ravel(), bins=nbins, orientation="horizontal"
                )
                # set color of histogram bins according to z value
                fracs = np.linspace(min_val, max_val, nbins)
                norm = colors.Normalize(min_val, max_val)
                for frac, patch in zip(fracs, patches):
                    color = getattr(cm, cmap)(norm(frac))
                    patch.set_facecolor(color)
                # make sliders to control cmin and cmax
                ax_min_slider = main_divider.append_axes("right", size="20%", pad=0.25)
                self.min_slider = min_slider = VertSlider(
                    ax_min_slider,
                    "min",
                    min_val,
                    max_val,
                    fontsize=10,
                    valinit=min_val,
                    labels=True,
                    alpha=1,
                    facecolor=getattr(cm, cmap)(norm(min_val)),
                )
                ax_max_slider = main_divider.append_axes("right", size="20%", pad=0.25)
                self.max_slider = max_slider = VertSlider(
                    ax_max_slider,
                    "max",
                    min_val,
                    max_val,
                    slidermin=min_slider,
                    fontsize=10,
                    valinit=max_val,
                    labels=True,
                    alpha=1,
                    facecolor=getattr(cm, cmap)(norm(max_val)),
                    start_at_bottom=False,
                )
                # function called when min_slider or max_slider is moved
                def update_cval(val):
                    cmin = min_slider.val
                    cmax = max_slider.val
                    cbar.mappable.set_clim([cmin, cmax])
                    im.set_clim([cmin, cmax])
                    upper.set_ydata(cmax)
                    lower.set_ydata(cmin)
                    min_slider.valmax = cmax
                    norm = colors.Normalize(cmin, cmax)
                    for frac, patch in zip(fracs, patches):
                        if cmin <= frac <= cmax:
                            color = getattr(cm, cmap)(norm(frac))
                            patch.set_alpha(1)
                        else:
                            color = "k"
                            patch.set_alpha(0.25)
                        patch.set_facecolor(color)

                for s in [min_slider, max_slider]:
                    s.on_changed(update_cval)
            # now add a subplot for the slice data
            ax1 = plt.subplot2grid((12, 12), (7, 2), colspan=8, rowspan=5, fig=self.fig)
            ax1.grid(self.get_opt("grid"))
            xlab = xlabel if slice_state == "x" else ylabel
            label = zlabel.split(" ")[:1] + ["".join(zlabel.split(" ")[1:])]
            ylab = "\n".join(label)
            color_by_value = self.line_color_btn.isChecked()
            if slice_state == "x":
                if color_by_value:
                    points = np.array(
                        [xs[1].magnitude, zs[1].magnitude[:, 0]]
                    ).T.reshape(-1, 1, 2)
                    # make segments overlap
                    segments = np.concatenate(
                        [points[:-2], points[1:-1], points[2:]], axis=1
                    )
                    lc = LineCollection(
                        segments,
                        cmap=cmap,
                        norm=colors.Normalize(*cbar.mappable.get_clim()),
                    )
                    lc.set_array(zs[1].magnitude[:, 0])
                    lc.set_linewidth(plot_lw)
                    ax1.add_collection(lc)
                    # we need an invisible line so that the figure draws correctly
                    (line,) = ax1.plot([0, 0], alpha=0)
                    xdata, ydata = xs[1].magnitude, zs[1].magnitude[:, 0]
                else:
                    (line,) = ax1.plot(
                        xs[1].magnitude, zs[1].magnitude[:, 0], lw=plot_lw
                    )
                    xdata, ydata = line.get_xdata(), line.get_ydata()
                cut = ax0.axhline(y=ax0.get_ylim()[0], color="k", alpha=0.8, lw=2)
            else:
                if color_by_value:
                    points = np.array(
                        [ys[1].magnitude, zs[1].magnitude[0, :]]
                    ).T.reshape(-1, 1, 2)
                    # make segments overlap
                    segments = np.concatenate(
                        [points[:-2], points[1:-1], points[2:]], axis=1
                    )
                    lc = LineCollection(
                        segments,
                        cmap=cmap,
                        norm=colors.Normalize(*cbar.mappable.get_clim()),
                    )
                    lc.set_array(zs[1].magnitude[0, :])
                    lc.set_linewidth(plot_lw)
                    ax1.add_collection(lc)
                    # we need an invisible line so that the figure draws correctly
                    (line,) = ax1.plot([0, 0], alpha=0)
                    xdata, ydata = ys[1].magnitude, zs[1].magnitude[0, :]
                else:
                    (line,) = ax1.plot(
                        ys[1].magnitude, zs[1].magnitude[0, :], lw=plot_lw
                    )
                    xdata, ydata = line.get_xdata(), line.get_ydata()
                cut = ax0.axvline(x=ax0.get_xlim()[0], color="k", alpha=0.8, lw=2)
            ax1.set_xlabel(xlab)
            ax1.set_ylabel(ylab)

            if self.line_color_btn.isChecked() and self.get_opt("histogram"):
                # adjust line color based on min_slider and max_slider
                def update_line_color(val):
                    lc.set_norm(colors.Normalize(min_slider.val, max_slider.val))

                for s in [min_slider, max_slider]:
                    s.on_changed(update_line_color)
            # add an axis for the slice slider
            divider = make_axes_locatable(ax1)
            ax_slider = divider.append_axes("bottom", size="15%", pad=0.45)
            idx_label = "y" if slice_state == "x" else "x"
            idx = 1 if slice_state == "x" else 0
            self.slider = slider = Slider(
                ax_slider,
                f"{idx_label} index",
                0,
                zs[1].shape[idx] - 1,
                valinit=0,
                valstep=1,
                valfmt="%i",
            )
            z = rotate(zs[1].magnitude.T, angle, cval=np.nan)
            x = np.linspace(*ax0.get_xlim(), z.shape[1])
            y = np.linspace(*ax0.get_ylim(), z.shape[0])
            # function called when slider is moved
            def update(val):
                i = int(slider.val)
                z = rotate(zs[1].magnitude.T, angle, cval=np.nan)
                x = np.linspace(*ax0.get_xlim(), z.shape[1])
                y = np.linspace(*ax0.get_ylim(), z.shape[0])
                margin = 0.025
                color_by_value = self.line_color_btn.isChecked()
                if slice_state == "x":
                    slider.valmax = len(y) - 1
                    if color_by_value:
                        points = np.array([x, z[:, i]]).T.reshape(-1, 1, 2)
                        # make segments overlap
                        segments = np.concatenate(
                            [points[:-2], points[1:-1], points[2:]], axis=1
                        )
                        lc.set_segments(segments)
                        lc.set_array(z[:, i])
                        lc.set_norm(colors.Normalize(*cbar.mappable.get_clim()))
                    else:
                        line.set_xdata(x)
                        line.set_ydata(z[:, i])
                    rng = np.nanmax(x) - np.nanmin(x)
                    xmin = np.nanmin(x) - margin * rng
                    xmax = np.nanmax(x) + margin * rng
                    cut.set_ydata(2 * [y[i]])
                    xdata, ydata = x, z[:, i]
                    self.exp_data["slice"] = {
                        xs[0]: {"array": xdata, "unit": str(xs[1].units)},
                        zs[0]: {"array": ydata, "unit": str(zs[1].units)},
                        "index": int(slider.val),
                    }
                elif slice_state == "y":
                    slider.valmax = len(x) - 1
                    if color_by_value:
                        points = np.array([y, z[i, :]]).T.reshape(-1, 1, 2)
                        # make segments overlap
                        segments = np.concatenate(
                            [points[:-2], points[1:-1], points[2:]], axis=1
                        )
                        lc.set_segments(segments)
                        lc.set_array(z[i, :])
                        lc.set_norm(colors.Normalize(*cbar.mappable.get_clim()))
                    else:
                        line.set_xdata(y)
                        line.set_ydata(z[i, :])
                    rng = np.nanmax(y) - np.nanmin(y)
                    xmin = np.nanmin(y) - margin * rng
                    xmax = np.nanmax(y) + margin * rng
                    cut.set_xdata(2 * [x[i]])
                    xdata, ydata = y, z[i, :]
                    self.exp_data["slice"] = {
                        ys[0]: {"array": xdata, "unit": str(ys[1].units)},
                        zs[0]: {"array": ydata, "unit": str(zs[1].units)},
                        "index": int(slider.val),
                    }
                ax1.set_xlim(xmin, xmax)
                slider.ax.set_xlim(slider.valmin, slider.valmax)
                vmin, vmax = np.nanmin(ydata), np.nanmax(ydata)
                margin = 0.1
                rng = vmax - vmin
                vmin = vmin - margin * rng
                vmax = vmax + margin * rng
                try:
                    ax1.set_ylim(vmin, vmax)
                except ValueError:  # vmin == vmax
                    pass
                self.canvas.draw()
                self.fig.tight_layout()

            update(0)
            slider.on_changed(update)
            self.exp_data = {
                xs[0]: {"array": x, "unit": str(xs[1].units)},
                ys[0]: {"array": y, "unit": str(ys[1].units)},
                zs[0]: {"array": z.T, "unit": str(zs[1].units)},
            }

    def replot(self):
        """Update the current plot from self.current_data."""
        if self.current_data is not None:
            xs, ys, zs = self.current_data[:]
            if self.xy_units_box.isChecked():
                try:
                    xs[1].ito(self.xy_units.text())
                except:
                    self.xy_units.setText(str(xs[1].units))
            if zs is None:
                name = ys[0]
                try:
                    ys[1].ito(self.units.text())
                except:
                    self.units.setText(str(ys[1].units))
            else:
                name = zs[0]
                if self.xy_units_box.isChecked():
                    try:
                        ys[1].ito(self.xy_units.text())
                    except:
                        self.xy_units.setText(str(ys[1].units))
                try:
                    zs[1].ito(self.units.text())
                except:
                    self.units.setText(str(zs[1].units))
            name = ys[0] if zs is None else zs[0]
            self.fig_title = f"{self.dataset.metadata['location']} [{name}]"
            self.current_data = [xs, ys, zs]
            self.subtract_background()

    def set_slice(self, idx=None, replot=True):
        """Capture the user-requested slice state."""
        if isinstance(idx, QtWidgets.QRadioButton):
            idx = self.slice_radio.id(idx)
        if idx is None:
            idx = self.slice_radio.checkedId()
        self.slice_state = idx
        self.update_slice()
        if replot:
            self.replot()

    def update_slice(self):
        """Show or hide pyqtgraph slice widgets based on self.slice_state."""
        if self.slice_state == 0:
            self.pyqt_imview.x_slice_widget.hide()
            self.pyqt_imview.y_slice_widget.hide()
            self.line_color_btn.setEnabled(False)
        elif self.slice_state == 1:
            self.pyqt_imview.x_slice_widget.show()
            self.pyqt_imview.y_slice_widget.hide()
            self.line_color_btn.setEnabled(True)
        elif self.slice_state == 2:
            self.pyqt_imview.x_slice_widget.hide()
            self.pyqt_imview.y_slice_widget.show()
            self.line_color_btn.setEnabled(True)
        else:
            raise ValueError("Unknown Slice State: {}".format(self.slice_state))

    def update_line_by_line(self):
        """Enable/disable line-by-line background subtraction based on self.line_backsub_btn."""
        self.x_line_backsub_btn.setEnabled(self.line_backsub_btn.isChecked())
        self.y_line_backsub_btn.setEnabled(self.line_backsub_btn.isChecked())
        self.replot()

    def subtract_background(self, idx=None):
        """Perform global or line-by-line background subtraction."""
        if self.current_data is None:
            return
        if isinstance(idx, QtWidgets.QRadioButton):
            idx = self.backsub_radio.id(idx)
        idx = idx or self.backsub_radio.checkedId()
        xs, ys, zs = self.current_data[:]  # copy self.current_data to avoid changing it
        line_by_line = self.line_backsub_btn.isChecked()
        if line_by_line and zs is not None:
            funcs = (
                lambda x: 0,
                np.min,
                np.max,
                np.mean,
                np.median,
                lambda y, x=xs[1].magnitude: self._subtract_line(x, y),
            )
            axis = self.line_backsub_radio.checkedId()
            z = self._subtract_line_by_line(np.copy(zs[1].magnitude), axis, funcs[idx])
            zs = [zs[0], z * zs[1].units]  # restore units after background subtraction
        if idx == 1:  # min
            if zs is None:
                ys = [ys[0], ys[1].units * (ys[1].m - np.min(ys[1].m))]
            elif not line_by_line:
                zs = [zs[0], zs[1].units * (zs[1].m - np.min(zs[1].m))]
        elif idx == 2:  # max
            if zs is None:
                ys = [ys[0], ys[1].units * (ys[1].m - np.max(ys[1].m))]
            elif not line_by_line:
                zs = [zs[0], zs[1].units * (zs[1].m - np.max(zs[1].m))]
        elif idx == 3:  # mean
            if zs is None:
                ys = [ys[0], ys[1].units * (ys[1].m - np.mean(ys[1].m))]
            elif not line_by_line:
                zs = [zs[0], zs[1].units * (zs[1].m - np.mean(zs[1].m))]
        elif idx == 4:  # median
            if zs is None:
                ys = [ys[0], ys[1] - ys[1].units * np.median(ys[1].m)]
            elif not line_by_line:
                zs = [
                    zs[0],
                    zs[1] - zs[1].units * np.median(np.reshape(zs[1].m, (-1, 1))),
                ]
        elif idx == 5:  # linear
            if zs is None:
                slope, offset = np.polyfit(xs[1].magnitude, ys[1].magnitude, 1)
                ys = [ys[0], ys[1] - ys[1].units * (slope * xs[1].magnitude + offset)]
            elif not line_by_line:
                X, Y = np.meshgrid(xs[1].m, ys[1].m, indexing="ij")
                x = np.reshape(X, (-1, 1))
                y = np.reshape(Y, (-1, 1))
                data = np.reshape(zs[1].m, (-1, 1))
                z = np.column_stack((x, y, np.ones_like(x)))
                plane, _, _, _ = lstsq(z, data)
                zs = [
                    zs[0],
                    zs[1] - zs[1].units * (plane[0] * X + plane[1] * Y + plane[2]),
                ]
        self.plot_arrays(xs, ys, zs=zs, title=self.fig_title)

    def export_mpl(self, dpi=220):
        """Open dialog to export matplotlib figure."""
        if self.dataset is None:
            return
        name = self.fig_title.split("/")[-1].replace(" ", "")
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export matplotlib", name, "PNG (*.png)"
        )
        print(f"Saving fig with dpi={dpi}")
        self.fig.savefig(path, dpi=dpi)

    def export_qt(self, width=1200):
        """Open dialog to export pyqtgraph figure.
        Args:
            width (int): Figure width in pixels (I think).
        Note: this is not currently used by the GUI because the pyqtgraph figures are ugly.
        """
        if self.dataset is None:
            return
        name = self.fig_title.split("/")[-1].replace(" ", "")
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export pyqtgraph", name, "PNG Image (*.png);; JPG Image (*.jpg)"
        )
        plot = (
            self.pyqt_plot.plotItem
            if self.pyqt_plot.isVisible()
            else self.pyqt_imview.scene
        )
        exporter = pg.exporters.ImageExporter(plot)
        exporter.parameters()["width"] = width
        exporter.export(path)

    def export_data(self):
        """Open dialog to export current data (post transformations and background subtraction)
        to .mat, .h5, or .pickle files.
        """
        if not self.exp_data:
            return
        name = self.fig_title.split("/")[-1].replace(" ", "")
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Export current data",
            name,
            "MAT (*.mat);;HDF5 (*.h5);;pickle (*.pickle)",
        )
        if path.endswith("mat"):
            try:
                savemat(path, self.exp_data)
            except:
                pass
        elif path.endswith("h5"):
            try:
                with h5py.File(path) as df:
                    set_h5_attrs(df, self.exp_data)
            except:
                pass
        elif path.endswith("pickle"):
            try:
                with open(path, "wb") as f:
                    pickle.dump(self.exp_data, f)
            except:
                pass

    def _subtract_line_by_line(self, zdata, axis, func):
        """Perform line-by-line background subtraction of `zdata` along axis `axis` according
        to callable `func`.
        Args:
            zdata (np.ndarray): 2D data for which you want to do background subtraction.
            axis (int): Axis along which you want to do line-by-line background subtraction.
            func (callable): Function applied to each line to calculate the value to subtract
                (e.g. np.min, np.mean, etc.)
        Returns:
            np.ndarray: zdata with background subtracted line-by-line.
        """
        if axis:  # y
            for i in range(zdata.shape[axis]):
                zdata[:, i] -= func(zdata[:, i])
        else:  # x
            for i in range(zdata.shape[axis]):
                zdata[i, :] -= func(zdata[i, :])
        return zdata

    def _subtract_line(self, x, y):
        """Subtract the best-fit line `slope * x + offset` from array `y`.
        Args:
            x (np.ndarray): OD or 1D array of x values.
            y (np.ndarray): 0D or 1D array of y values.
        Returns:
            np.ndarray: y with best-fit line subtracted.
        """
        slope, offset = np.polyfit(x, y, 1)
        return y - (slope * x + offset)


class DataSetPlotter(PlotWidget):
    """PlotWidget with functionality specific to scanning-squid datasets."""

    def __init__(self, dataset=None, parent=None):
        super().__init__(parent=parent)
        self.dataset = dataset
        self.arrays = None
        self.indep_vars = None
        arrays_widget = QtWidgets.QGroupBox("Arrays")
        arrays_layout = QtWidgets.QGridLayout(arrays_widget)
        arrays_widget.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Minimum
        )
        self.selector = QtWidgets.QComboBox()
        arrays_layout.addWidget(self.selector, 0, 0)
        self.option_layout.insertWidget(0, arrays_widget)
        self.selector.currentIndexChanged.connect(self.set_plot)
        self.units = QtWidgets.QLineEdit("Array unit")
        self.units.setEnabled(False)
        self.units.returnPressed.connect(self.replot)
        arrays_layout.addWidget(self.units, 1, 0)

        self.xy_units_box = QtWidgets.QCheckBox("real x-y units")
        self.xy_units_box.setChecked(False)
        self.xy_units = QtWidgets.QLineEdit()
        self.xy_units.setText("Length unit")
        self.xy_units.setEnabled(False)
        arrays_layout.addWidget(self.xy_units_box, 0, 1)
        arrays_layout.addWidget(self.xy_units, 1, 1)
        # xy_units_widget = QtWidgets.QWidget()
        # xy_units_layout = QtWidgets.QVBoxLayout()
        # xy_units_widget.setLayout(xy_units_layout)
        # xy_units_layout.addWidget(self.xy_units_box)
        # xy_units_layout.addWidget(self.xy_units)
        # arrays_layout.addWidget(xy_units_widget)
        self.xy_units_box.stateChanged.connect(self.update_xy_units)
        self.xy_units.returnPressed.connect(self.replot)

    def get_arrays(self):
        """Create a dict of arrays from self.dataset."""
        if self.dataset is None:
            return
        if "_scan_" in self.dataset.location:
            if self.xy_units_box.isChecked():
                try:
                    self.arrays = scan_to_arrays(
                        self.dataset, xy_unit=self.xy_units.text()
                    )
                except:
                    self.arrays = scan_to_arrays(self.dataset, xy_unit="um")
                    self.xy_units.setText("um")
            else:
                self.arrays = scan_to_arrays(self.dataset)
            self.indep_vars = ("x", "y")
        elif "_td_cap_" in self.dataset.location:
            if self.xy_units_box.isChecked():
                try:
                    self.arrays = td_to_arrays(
                        self.dataset, z_unit=self.xy_units.text()
                    )
                except:
                    self.arrays = td_to_arrays(self.dataset, z_unit="um")
                    self.xy_units.setText("um")
            else:
                self.arrays = td_to_arrays(self.dataset)
            self.indep_vars = ("height",)
        old_text = self.selector.currentText()
        old_units = self.units.text()
        self.selector.clear()
        for name in self.arrays.keys():
            if name.lower() not in self.indep_vars:
                self.selector.addItem(name)
        if old_text and old_text in self.arrays.keys():
            self.selector.setCurrentText(old_text)
            self.set_plot_from_name(old_text)
            if "Array" not in old_units:
                self.units.setText(old_units)
        else:
            self.selector.setCurrentIndex(0)
            self.set_plot_from_name(self.selector.currentText())

    def set_plot(self, idx):
        """Set current plot to the given index of self.selector.
        Args:
            idx (int): Index of requested plot.
        """
        name = str(self.selector.itemText(idx))
        if not name:
            return
        self.set_plot_from_name(name)

    def set_plot_from_name(self, name):
        """Set current plot by name.
        Args:
            name (str): Name of requested plot/array.
        """
        if len(self.indep_vars) == 1:
            xs = [self.indep_vars[0], self.arrays[self.indep_vars[0]]]
            ys = [name, self.arrays[name]]
            zs = None
            try:
                unit = self.units.text()
                ys[1].ito(unit)
            except:
                unit = ys[1].units
            self.units.setText(str(unit))
            self.units.setEnabled(True)
        elif len(self.indep_vars) == 2:
            xs, ys = ([var, self.arrays[var]] for var in self.indep_vars)
            z = self.arrays[name]
            # set nan values to the non-nan min to avoid issues with plotting
            z[np.isnan(z.magnitude)] = np.nanmin(z.magnitude) * z.units
            zs = [name, z]
            try:
                unit = self.units.text()
                zs[1].ito(unit)
            except:
                unit = zs[1].units
            self.units.setText(str(unit))
            self.units.setEnabled(True)
        title = ""
        if self.dataset is not None:
            title = f"{self.dataset.metadata['location']} [{name}]"
        self.current_data = [xs, ys, zs]
        self.plot_arrays(xs, ys, zs, title)
        self.subtract_background()

    def update_xy_units(self):
        """Enable or disable user entry for x-y length units."""
        if self.xy_units_box.isChecked():
            self.xy_units.setText("um")
            self.xy_units.setEnabled(True)
        else:
            self.xy_units.setText("Length unit")
            self.xy_units.setEnabled(False)
        self.update()

    def update(self, dataset=None):
        """Update arrays and plot from dataset."""
        self.dataset = dataset or self.dataset
        self.get_arrays()
        self.backsub_radio.button(0).setChecked(True)
        self.set_plot_from_name(self.selector.currentText())


class ImageView(pg.ImageView):
    """pyqtgraph ImageView wrapper."""

    def __init__(self, **kwargs):
        kwargs["view"] = pg.PlotItem(labels=kwargs.pop("labels", None))
        super().__init__(**kwargs)
        self.view.setAspectLocked(lock=True)
        self.view.invertY(False)
        self.set_histogram(True)
        histogram_action = QtWidgets.QAction("Histogram", self)
        histogram_action.setCheckable(True)
        histogram_action.triggered.connect(self.set_histogram)
        self.scene.contextMenu.append(histogram_action)
        self.ui.histogram.gradient.loadPreset("grey")

    def set_histogram(self, visible):
        """Show or hide the histogram.
        Args:
            visible (bool): Whether you want the histogram to be visible.
        """
        self.ui.histogram.setVisible(visible)
        self.ui.roiBtn.setVisible(False)
        self.ui.normGroup.setVisible(False)
        self.ui.menuBtn.setVisible(False)

    def set_data(self, data):
        """Set the image data.
        Args:
            data (np.ndarray): 2D array of iamge data.
        """
        self.setImage(data)

    def set_cmap(self, name):
        """Set the colormap to one of pyqtgraph's presets.
        Args:
            name (str): Name of preset colormap.
        """
        self.ui.histogram.gradient.loadPreset(name)


class SlicePlotWidget(pg.PlotWidget):
    """pyqtgraph PlotWidget with crosshairs that follow mouse."""

    crosshair_moved = QtCore.Signal(float, float)

    def __init__(self, parametric=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cross_section_enabled = False
        self.parametric = parametric
        self.search_mode = True
        self.label = None
        self.selected_point = None
        self.plotItem.showGrid(x=True, y=True)
        self.scene().sigMouseClicked.connect(self.toggle_search)
        self.scene().sigMouseMoved.connect(self.handle_mouse_move)

    def set_data(self, data, **kwargs):
        if data is not None and len(data) > 0 and np.isfinite(data).all():
            self.clear()
            self.plot(data, **kwargs)

    def toggle_search(self, mouse_event):
        """Toggle the crosshairs tracking mouse movement on click event."""
        if mouse_event.double():
            if self.cross_section_enabled:
                self.hide_crosshair()
            else:
                self.add_crosshair()
        elif self.cross_section_enabled:
            self.search_mode = not self.search_mode
            if self.search_mode:
                self.handle_mouse_move(mouse_event.scenePos())

    def handle_mouse_move(self, mouse_event):
        """Depending on search_mode and cross_section_enabled, track mouse movement
        and emit a signal with the crosshair position.
        """
        if self.cross_section_enabled and self.search_mode:
            item = self.getPlotItem()
            view_coords = item.getViewBox().mapSceneToView(mouse_event)
            view_x, view_y = view_coords.x(), view_coords.y()
            # try to get data indices corresponding to mouse position
            guesses = []
            for data_item in item.items:
                if isinstance(data_item, pg.PlotDataItem):
                    xdata, ydata = data_item.xData, data_item.yData
                    index_distance = (
                        lambda i: (xdata[i] - view_x) ** 2 + (ydata[i] - view_y) ** 2
                    )
                    if self.parametric:
                        index = min(range(len(xdata)), key=index_distance)
                    else:
                        index = min(np.searchsorted(xdata, view_x), len(xdata) - 1)
                        if index and xdata[index] - view_x > view_x - xdata[index - 1]:
                            index -= 1
                    pt_x, pt_y = xdata[index], ydata[index]
                    guesses.append(((pt_x, pt_y), index_distance(index), index))
            if not guesses:
                return

            (pt_x, pt_y), _, index = min(guesses, key=lambda x: x[1])
            self.selected_point = (pt_x, pt_y)
            self.v_line.setPos(pt_x)
            self.h_line.setPos(pt_y)
            self.label.setText("x={:.5f}, y={:.5f}".format(pt_x, pt_y))
            self.crosshair_moved.emit(pt_x, pt_y)

    def add_crosshair(self):
        self.h_line = pg.InfiniteLine(angle=0, movable=False)
        self.v_line = pg.InfiniteLine(angle=90, movable=False)
        self.addItem(self.h_line, ignoreBounds=False)
        self.addItem(self.v_line, ignoreBounds=False)
        if self.label is None:
            self.label = pg.LabelItem(justify="right")
            self.getPlotItem().layout.addItem(self.label, 4, 1)
        self.x_cross_index = 0
        self.y_cross_index = 0
        self.cross_section_enabled = True

    def hide_crosshair(self):
        self.removeItem(self.h_line)
        self.removeItem(self.v_line)
        self.cross_section_enabled = False


class SliceableImageView(ImageView):
    """ImageView combined with a SlicePlotWidget for both x and y slices."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.search_mode = False
        self._connect_signals()
        self.angle = 0
        self.y_cross_index = 0
        self.x_cross_index = 0
        self.x_slice_widget = SlicePlotWidget()
        self.x_slice_widget.add_crosshair()
        self.x_slice_widget.search_mode = False
        self.pen = pg.mkPen(width=plot_lw)
        self.x_slice_widget_data = self.x_slice_widget.plot([0, 0], pen=self.pen)
        self.h_line = pg.InfiniteLine(pos=0, angle=0, movable=False, pen=self.pen)
        self.view.addItem(self.h_line, ignoreBounds=False)

        self.y_slice_widget = SlicePlotWidget()
        self.y_slice_widget.add_crosshair()
        self.y_slice_widget.search_mode = False
        self.y_slice_widget_data = self.y_slice_widget.plot([0, 0], pen=self.pen)
        self.v_line = pg.InfiniteLine(pos=0, angle=90, movable=False, pen=self.pen)
        self.view.addItem(self.v_line, ignoreBounds=False)

        self.x_slice_widget.crosshair_moved.connect(lambda x, _: self.set_position(x=x))
        self.y_slice_widget.crosshair_moved.connect(lambda y, _: self.set_position(y=y))

        self.text_item = pg.LabelItem(justify="right")
        self.view.layout.addItem(self.text_item, 4, 1)

    def setImage(self, *args, **kwargs):
        """Set the image and adjust ViewBox, etc.
        *args and **kwargs passed to ImageItem constructor.
        """
        if "pos" in kwargs:
            self._x0, self._y0 = kwargs["pos"]
        else:
            self._x0, self._y0 = 0, 0
        if "scale" in kwargs:
            self._xscale, self._yscale = kwargs["scale"]
        else:
            self._xscale, self._yscale = 1, 1
        # adjust to make pixel centers align on ticks
        self._x0 -= self._xscale / 2.0
        self._y0 -= self._yscale / 2.0
        if "pos" in kwargs:
            kwargs["pos"] = self._x0, self._y0

        if self.imageItem.image is not None:
            (min_x, max_x), (min_y, max_y) = self.imageItem.getViewBox().viewRange()
            mid_x, mid_y = (max_x + min_x) / 2.0, (max_y + min_y) / 2.0
        else:
            mid_x, mid_y = 0, 0

        self.h_line.setPos(mid_y)
        self.v_line.setPos(mid_x)
        super().setImage(*args, **kwargs)
        self.set_position()

    def setLabels(self, xlabel="x", ylabel="y", zlabel="z"):
        """Set x, y, and z labels."""
        self.view.setLabels(bottom=(xlabel,), left=(ylabel,))
        self.x_slice_widget.plotItem.setLabels(bottom=xlabel, left=zlabel)
        self.y_slice_widget.plotItem.setLabels(bottom=ylabel, left=zlabel)
        self.ui.histogram.item.axis.setLabel(text=zlabel)

    def _connect_signals(self):
        """Setup signals."""
        if self.imageItem.scene() is None:
            raise RuntimeError(
                "Signal can only be connected after it has been embedded in a scene."
            )
        self.imageItem.scene().sigMouseClicked.connect(self.toggle_search)
        self.imageItem.scene().sigMouseMoved.connect(self.handle_mouse_move)
        self.timeLine.sigPositionChanged.connect(self.update_slice)

    def toggle_search(self, mouse_event):
        """Toggle the crosshairs tracking mouse movement on click event."""
        if mouse_event.double():
            return
        self.search_mode = not self.search_mode
        if self.search_mode:
            self.handle_mouse_move(mouse_event.scenePos())

    def handle_mouse_move(self, mouse_event):
        """Depending on search_mode, track mouse movement and update position text_item."""
        if self.search_mode:
            view_coords = self.imageItem.getViewBox().mapSceneToView(mouse_event)
            view_x, view_y = view_coords.x(), view_coords.y()
            self.set_position(view_x, view_y)

    def set_position(self, x=None, y=None):
        """Update text_item displaying x, y, and z mouse position."""
        if x is None:
            x = self.v_line.getXPos()
        if y is None:
            y = self.h_line.getYPos()

        item_coords = self.imageItem.getViewBox().mapFromViewToItem(
            self.imageItem, QtCore.QPointF(x, y)
        )
        item_x, item_y = item_coords.x(), item_coords.y()
        max_x, max_y = self.imageItem.image.shape

        item_x = self.x_cross_index = max(min(int(item_x), max_x - 1), 0)
        item_y = self.y_cross_index = max(min(int(item_y), max_y - 1), 0)

        view_coords = self.imageItem.getViewBox().mapFromItemToView(
            self.imageItem, QtCore.QPointF(item_x + 0.5, item_y + 0.5)
        )
        x, y = view_coords.x(), view_coords.y()

        self.v_line.setPos(x)
        self.h_line.setPos(y)
        z_val = self.imageItem.image[self.x_cross_index, self.y_cross_index]
        self.update_slice()
        self.text_item.setText("x={:.5f}, y={:.5f}, z={:.5f}".format(x, y, z_val))

    def update_slice(self):
        """Update the current x and y slices."""
        zdata = self.imageItem.image
        nx, ny = zdata.shape
        x0, y0, xscale, yscale = self._x0, self._y0, self._xscale, self._yscale
        xdata = np.linspace(x0, x0 + (xscale * (nx - 1)), nx)
        ydata = np.linspace(y0, y0 + (yscale * (ny - 1)), ny)
        zval = zdata[self.x_cross_index, self.y_cross_index]
        self.x_slice_widget_data.setData(xdata, zdata[:, self.y_cross_index])
        self.x_slice_widget.v_line.setPos(xdata[self.x_cross_index])
        self.x_slice_widget.h_line.setPos(zval)
        self.y_slice_widget_data.setData(ydata, zdata[self.x_cross_index, :])
        self.y_slice_widget.v_line.setPos(ydata[self.y_cross_index])
        self.y_slice_widget.h_line.setPos(zval)



from ..qjsonmodel import QJsonModel
from ..utils import load_json_ordered
from ..qt import *


class MetaWidget(QtWidgets.QTreeView):
    """Widget for inspecting dataset metadata."""

    def __init__(self, path=None, data=None):
        super(MetaWidget, self).__init__()
        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.model = QJsonModel()
        self.setModel(self.model)
        assert not (path and data), "path and data cannot both be None."
        if path is not None:
            self.load_meta(path=path)
        elif data is not None:
            self.load_meta(data=data)
        self.expanded.connect(lambda item: self.resizeColumnToContents(0))
        self.setHeaderHidden(True)

    def load_meta(self, path=None, data=None):
        """Loads metadata stored in json file at `path`."""
        assert not (path and data), "path and data cannot both be None."
        if path is not None:
            self.model.load(load_json_ordered(path))
        elif data is not None:
            self.model.load(data)
        self.resizeColumnToContents(0)



from qtpy import QtGui, QtCore, QtWidgets, API
from qtpy.QtCore import Qt

if API == "pyqt5":
    from matplotlib.backends.backend_qt5agg import (
        FigureCanvasQTAgg,
        NavigationToolbar2QT,
    )
else:
    from matplotlib.backends.backend_qt4agg import (
        FigureCanvasQTAgg,
        NavigationToolbar2QT,
    )

if hasattr(Qt, "AA_EnableHighDpiScaling"):
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)

if hasattr(Qt, "AA_UseHighDpiPixmaps"):
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

__all__ = [
    "QtGui",
    "QtCore",
    "QtWidgets",
    "Qt",
    "FigureCanvasQTAgg",
    "NavigationToolbar2QT",
]



import os
from ..qt import *
import qcodes as qc
from qcodes.data.data_set import load_data
from ..qjsonmodel import QJsonModel
from .plots import DataSetPlotter
from ..utils import load_json_ordered


class DataSetBrowser(QtWidgets.QWidget):
    """Widget for selecting dates and datasets from a directory."""

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.directory = None
        self.date_selector = QtWidgets.QComboBox(parent=self)
        self.dataset_selector = QtWidgets.QListView(parent=self)
        self.dataset_list = QtGui.QStandardItemModel(parent=self)
        self.dataset_selector.setModel(self.dataset_list)
        self.dataset_selector.setIconSize(QtCore.QSize(150, 150))

        button_widget = QtWidgets.QWidget()
        button_layout = QtWidgets.QHBoxLayout()
        button_widget.setLayout(button_layout)
        select_dir_button = QtWidgets.QPushButton("Select directory")
        select_dir_button.clicked.connect(self.select_from_dialog)
        refresh_button = QtWidgets.QPushButton("Refresh")
        button_layout.addWidget(select_dir_button)
        button_layout.addWidget(refresh_button)
        selector_layout = QtWidgets.QVBoxLayout(self)
        selector_layout.addWidget(button_widget)
        selector_layout.addWidget(self.date_selector)
        selector_layout.addWidget(self.dataset_selector)

        self.date_selector.currentIndexChanged.connect(self.set_available_datasets)
        refresh_button.clicked.connect(self.set_available_dates)

    def select_from_dialog(self):
        """Update self.directory based on directory selected by user."""
        self.directory = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Select data directory"
        )
        self.set_available_dates()

    def set_available_dates(self):
        """Populate self.date_selector combo box with available date directories in self.directory."""
        self.date_selector.clear()
        if not os.path.isdir(self.directory):
            return
        for item in sorted(os.listdir(self.directory), reverse=True):
            if item.startswith("."):
                continue
            path = os.path.join(self.directory, item)
            if os.path.isdir(path):
                self.date_selector.addItem(item)

    def get_dataset(self):
        """Try to load the current selected dataset. If it can't be loaded, returns None."""
        self.dataset_name = self.dataset_selector.currentIndex().data()
        if not self.dataset_name or not self.date_name:
            return None
        fpath = os.path.join(self.directory, self.date_name, self.dataset_name)
        return load_data(location=fpath)

    def set_available_datasets(self, index):
        """Populates self.dataset_list with the available datasets for the currently selected date."""
        self.dataset_list.clear()
        self.date_name = str(self.date_selector.currentText())
        if not self.date_name:
            return None
        fpath = os.path.join(self.directory, self.date_name)
        items = sorted(
            [d for d in os.listdir(fpath) if os.path.isdir(os.path.join(fpath, d))],
            reverse=True,
        )

        for d in items:
            item = QtGui.QStandardItem(d)
            item.setEditable(False)
            image_list = [
                f for f in os.listdir(os.path.join(fpath, d)) if f.endswith(".png")
            ]
            if image_list:
                icon_path = os.path.join(fpath, d, image_list[0])
                item.setIcon(QtGui.QIcon(icon_path))
            self.dataset_list.appendRow(item)



"""Analysis GUI for scanning SQUID microscopy datasets generated by the
[scanning-squid](https://github.com/moler-group/scanning-squid) python package.
"""

from setuptools import setup, find_packages

DESCRIPTION = "Analysis GUI for scanning SQUID microscopy"
LONG_DESCRIPTION = __doc__

NAME = "ssm_analyze"
AUTHOR = "Logan Bishop-Van Horn"
AUTHOR_EMAIL = "logan.bvh@gmail.com"
URL = "https://github.com/loganbvh/ssm_analyze"
LICENSE = "MIT"
PYTHON_VERSION = ">=3.6"

INSTALL_REQUIRES = [
    "h5py",
    "ipykernel",
    "ipython",
    "jupyter",
    "jupyter_console",
    "matplotlib",
    "numpy",
    "pint",
    "PyQt5",
    "pyqtgraph",
    "qcodes",
    "qtpy",
    "scipy",
]

CLASSIFIERS = """\
Development Status :: 4 - Beta
Intended Audience :: Science/Research
License :: OSI Approved :: MIT License
Operating System :: MacOS
Operating System :: POSIX
Operating System :: Unix
Operating System :: Microsoft :: Windows
Programming Language :: Python
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Physics
"""

CLASSIFIERS = [line for line in CLASSIFIERS.splitlines() if line]
PLATFORMS = ["Linux", "Mac OSX", "Unix", "Windows"]
KEYWORDS = "scanning SQUID microscopy"


setup(
    name=NAME,
    version="0.2.3",
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    license=LICENSE,
    packages=find_packages(),
    package_data={
        "ssm_analyze": ["gui/img/icon.png", "sample_data/*"],
    },
    include_package_data=True,
    entry_points={
        "console_scripts": ["ssm_analyze=ssm_analyze.gui.window:main"],
    },
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    platforms=PLATFORMS,
    python_requires=PYTHON_VERSION,
    install_requires=INSTALL_REQUIRES,
)



from ..qt import Qt, QtCore, QtWidgets, QtGui
from matplotlib.widgets import AxesWidget


class LabeledSlider(QtWidgets.QWidget):
    """QSlider with ticks and labels."""

    def __init__(
        self,
        minimum,
        maximum,
        init,
        interval=1,
        orientation=Qt.Horizontal,
        labels=None,
        parent=None,
    ):
        super().__init__(parent=parent)

        levels = range(minimum, maximum + interval, interval)
        if labels is not None:
            if not isinstance(labels, (tuple, list)):
                raise Exception("<labels> is a list or tuple.")
            if len(labels) != len(levels):
                raise Exception("Size of <labels> doesn't match levels.")
            self.levels = list(zip(levels, labels))
        else:
            self.levels = list(zip(levels, map(str, levels)))

        if orientation == Qt.Horizontal:
            self.layout = QtWidgets.QVBoxLayout(self)
        elif orientation == Qt.Vertical:
            self.layout = QtWidgets.QHBoxLayout(self)
        else:
            raise Exception("<orientation> wrong.")

        # gives some space to print labels
        self.left_margin = 10
        self.top_margin = 0
        self.right_margin = 10
        self.bottom_margin = 10

        self.layout.setContentsMargins(
            self.left_margin, self.top_margin, self.right_margin, self.bottom_margin
        )

        self.slider = QtWidgets.QSlider(orientation, self)
        self.slider.setMinimum(minimum)
        self.slider.setMaximum(maximum)
        self.slider.setValue(init)
        if orientation == Qt.Horizontal:
            self.slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        else:
            self.slider.setTickPosition(QtWidgets.QSlider.TicksLeft)
        self.slider.setTickInterval(interval)
        self.slider.setSingleStep(1)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Minimum
        )

        self.layout.addWidget(self.slider)

    def paintEvent(self, e):
        super().paintEvent(e)
        style = self.slider.style()
        painter = QtGui.QPainter(self)
        painter.setPen(QtGui.QPen(Qt.black, 1))
        st_slider = QtWidgets.QStyleOptionSlider()
        st_slider.initFrom(self.slider)
        st_slider.orientation = self.slider.orientation()

        length = style.pixelMetric(
            QtWidgets.QStyle.PM_SliderLength, st_slider, self.slider
        )
        available = style.pixelMetric(
            QtWidgets.QStyle.PM_SliderSpaceAvailable, st_slider, self.slider
        )

        for v, v_str in self.levels:
            # get the size of the label
            rect = painter.drawText(QtCore.QRect(), Qt.TextDontPrint, v_str)

            if self.slider.orientation() == Qt.Horizontal:
                # I assume the offset is half the length of slider, therefore
                # + length//2
                x_loc = (
                    QtWidgets.QStyle.sliderPositionFromValue(
                        self.slider.minimum(), self.slider.maximum(), v, available
                    )
                    + length // 2
                )

                # left bound of the text = center - half of text width + L_margin
                left = x_loc - rect.width() // 2 + self.left_margin
                bottom = self.rect().bottom()

                # enlarge margins if clipping
                if v == self.slider.minimum():
                    if left <= 0:
                        self.left_margin = rect.width() // 2 - x_loc
                    if self.bottom_margin <= rect.height():
                        self.bottom_margin = rect.height()

                    self.layout.setContentsMargins(
                        self.left_margin,
                        self.top_margin,
                        self.right_margin,
                        self.bottom_margin,
                    )

                if (
                    v == self.slider.maximum()
                    and rect.width() // 2 >= self.right_margin
                ):
                    self.right_margin = rect.width() // 2
                    self.layout.setContentsMargins(
                        self.left_margin,
                        self.top_margin,
                        self.right_margin,
                        self.bottom_margin,
                    )

            else:
                y_loc = QtWidgets.QStyle.sliderPositionFromValue(
                    self.slider.minimum(),
                    self.slider.maximum(),
                    v,
                    available,
                    upsideDown=True,
                )

                bottom = y_loc + length // 2 + rect.height() // 2 + self.top_margin - 3
                # there is a 3 px offset that I can't attribute to any metric

                left = self.left_margin - rect.width()
                if left <= 0:
                    self.left_margin = rect.width() + 2
                    self.layout.setContentsMargins(
                        self.left_margin,
                        self.top_margin,
                        self.right_margin,
                        self.bottom_margin,
                    )

            pos = QtCore.QPoint(left, bottom)
            painter.drawText(pos, v_str)
        return


class SliderWidget(QtWidgets.QWidget):
    """LabeledSlider synced with a spinbox to update/display current slider value."""

    def __init__(self, min, max, init, interval, parent=None):
        super().__init__(parent=parent)
        self.value_box = QtWidgets.QSpinBox(self)
        self.slider = LabeledSlider(min, max, init, interval=interval)
        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(self.value_box)
        layout.addWidget(self.slider)
        self.slider.slider.valueChanged.connect(self.update_value)
        self.value_box.setMinimum(min)
        self.value_box.setMaximum(max)
        self.value_box.setValue(init)
        self.value_box.setKeyboardTracking(False)
        self.value_box.valueChanged.connect(self.update_value)
        self.update_value(init)
        sp = self.sizePolicy()
        sp.setVerticalPolicy(QtWidgets.QSizePolicy.Minimum)
        sp.setHorizontalPolicy(QtWidgets.QSizePolicy.MinimumExpanding)
        self.setSizePolicy(sp)

    def update_value(self, val):
        self.value_box.setValue(val)
        self.slider.slider.setValue(val)


class VertSlider(AxesWidget):
    """
    A slider representing a floating point range.
    For the slider to remain responsive you must maintain a
    reference to it.
    """

    def __init__(
        self,
        ax,
        label,
        valmin,
        valmax,
        valinit=0,
        valfmt="%1.2f",
        closedmin=True,
        closedmax=True,
        slidermin=None,
        slidermax=None,
        dragging=True,
        fontsize=None,
        labels=True,
        start_at_bottom=True,
        **kwargs
    ):
        """
        Create a slider from *valmin* to *valmax* in axes *ax*.

        Additional kwargs are passed on to ``self.poly`` which is the
        :class:`matplotlib.patches.Rectangle` which draws the slider
        knob.  See the :class:`matplotlib.patches.Rectangle` documentation
        valid property names (e.g., *facecolor*, *edgecolor*, *alpha*, ...).

        Args:
            ax (Axes): The Axes to put the slider in
            label (str): Slider label
            valmin (float): The minimum value of the slider
            valmax (float): The maximum value of the slider
            valinit (optional, float): The slider initial position. Default: 0
            valfmt (str): Used to format the slider value, fprint format string
            closedmin (optional, bool): Indicate whether the slider interval is closed on the bottom. Default: True
            closedmax (optional, bool): Indicate whether the slider interval is closed on the top. Default: True
            slidermin (optional, Slider) Do not allow the current slider to have a value less than
                `slidermin`. Default: None
            slidermax (optional, Slider): Do not allow the current slider to have a value greater than
                `slidermax`. Default: None
            dragging (optional, bool): Whether the slider can be dragged by the mouse. Default: True.
            fontsize (optional, int): If provided, sets the label font size. Default: None
            labels (optional, bool): Sets whether to show labels. Default: True
            start_at_bottom (optional, bool): Sets whether the filled rectangle starts at the lower or upper
                end of the axis. Default: True
            **kwargs: Passed to self.poly, see above.

        """
        super().__init__(ax)
        self.valmin = valmin
        self.valmax = valmax
        self.val = valinit
        self.valinit = valinit
        if start_at_bottom:
            self.poly = ax.axhspan(valmin, valinit, 0, 1, **kwargs)
        else:
            self.poly = ax.axhspan(valinit, valmax, 0, 1, **kwargs)

        self.hline = ax.axhline(valinit, 0, 1, color="k", lw=2)

        self.valfmt = valfmt
        ax.set_xticks([])
        ax.set_ylim((valmin, valmax))
        ax.set_yticks([])
        ax.set_navigate(False)

        self.connect_event("button_press_event", self._update)
        self.connect_event("button_release_event", self._update)
        if dragging:
            self.connect_event("motion_notify_event", self._update)
        if labels:
            self.label = ax.text(
                0.5,
                1.05,
                label,
                transform=ax.transAxes,
                verticalalignment="center",
                horizontalalignment="center",
                size=fontsize,
            )

            self.valtext = ax.text(
                0.5,
                -0.1,
                valfmt % valinit,
                transform=ax.transAxes,
                verticalalignment="center",
                horizontalalignment="center",
                size=fontsize,
            )

        self.cnt = 0
        self.observers = {}

        self.closedmin = closedmin
        self.closedmax = closedmax
        self.slidermin = slidermin
        self.slidermax = slidermax
        self.drag_active = False

    def _update(self, event):
        """Update the slider position."""
        if self.ignore(event):
            return

        if event.button != 1:
            return

        if event.name == "button_press_event" and event.inaxes == self.ax:
            self.drag_active = True
            event.canvas.grab_mouse(self.ax)

        if not self.drag_active:
            return

        elif (event.name == "button_release_event") or (
            event.name == "button_press_event" and event.inaxes != self.ax
        ):
            self.drag_active = False
            event.canvas.release_mouse(self.ax)
            return

        val = event.ydata
        if val <= self.valmin:
            if not self.closedmin:
                return
            val = self.valmin
        elif val >= self.valmax:
            if not self.closedmax:
                return
            val = self.valmax

        if self.slidermin is not None and val <= self.slidermin.val:
            if not self.closedmin:
                return
            val = self.slidermin.val

        if self.slidermax is not None and val >= self.slidermax.val:
            if not self.closedmax:
                return
            val = self.slidermax.val

        self.set_val(val)

    def set_val(self, val):
        """Set the slider value and update rectangle."""
        xy = self.poly.xy
        xy[1] = 0, val
        xy[2] = 1, val
        self.poly.xy = xy
        self.hline.set_ydata(val)
        if hasattr(self, "valtext"):
            self.valtext.set_text(self.valfmt % val)
        if self.drawon:
            self.ax.figure.canvas.draw_idle()
        self.val = val
        if not self.eventson:
            return
        for cid, func in self.observers.items():
            func(val)

    def on_changed(self, func):
        """When the slider value is changed, call *func* with the new
        slider position.

        A connection id is returned which can be used to disconnect
        """
        cid = self.cnt
        self.observers[cid] = func
        self.cnt += 1
        return cid

    def disconnect(self, cid):
        """Remove the observer with connection id *cid*."""
        try:
            del self.observers[cid]
        except KeyError:
            pass

    def reset(self):
        """Reset the slider to the initial value if needed."""
        if self.val != self.valinit:
            self.set_val(self.valinit)
